@@ -5,7 +5,7 @@
  *
  * Copyright (C) 2005-2006  NTT DATA CORPORATION
  *
- * Version: 1.1.2   2006/06/02
+ * Version: 1.1   2006/04/01
  *
  * This file is applicable to both 2.4.30 and 2.6.11 and later.
  * See README.ccs for ChangeLog.
@@ -17,6 +17,8 @@
 #include <linux/sakura.h>
 #include <linux/realpath.h>
 
+extern const char *ccs_log_level;
+
 /***** The structure for reserved ports. *****/
 
 typedef struct reserved_entry {
@@ -30,28 +32,27 @@ typedef struct reserved_entry {
 
 static RESERVED_ENTRY reservedport_list = { NULL, 0, 0, 0 };
 
-static int AddReservedEntry(const unsigned short int min_port, const unsigned short int max_port, const int is_delete)
+static int AddReservedEntry(const unsigned short int min_port, const unsigned short int max_port)
 {
 	RESERVED_ENTRY *new_entry, *ptr;
-	if (!is_delete) {
-		static spinlock_t lock = SPIN_LOCK_UNLOCKED;
-		/* I don't want to add if it was already added. */
-		for (ptr = reservedport_list.next; ptr; ptr = ptr->next) {
-			if (ptr->min_port == min_port && max_port == ptr->max_port) { ptr->is_deleted = 0; return 0; }
-		}
-		if ((new_entry = (RESERVED_ENTRY *) alloc_element(sizeof(RESERVED_ENTRY))) == NULL) return -ENOMEM;
-		new_entry->min_port = min_port;
-		new_entry->max_port = max_port;
-		/***** CRITICAL SECTION START *****/
-		spin_lock(&lock);
-		for (ptr = &reservedport_list; ptr->next; ptr = ptr->next); ptr->next = new_entry;
-		spin_unlock(&lock);
-		/***** CRITICAL SECTION END *****/
-	} else {
-		for (ptr = reservedport_list.next; ptr; ptr = ptr->next) {
-			if (ptr->min_port == min_port && max_port == ptr->max_port) ptr->is_deleted = 1;
-		}
+	static spinlock_t lock = SPIN_LOCK_UNLOCKED;
+	/* I don't want to add if it was already added. */
+	for (ptr = reservedport_list.next; ptr; ptr = ptr->next) {
+		if (ptr->min_port <= min_port && max_port <= ptr->max_port && !ptr->is_deleted) return 0;
 	}
+	if ((new_entry = (RESERVED_ENTRY *) alloc_element(sizeof(RESERVED_ENTRY))) == NULL) return -ENOMEM;
+	memset(new_entry, 0, sizeof(RESERVED_ENTRY));
+	new_entry->next = NULL;
+	new_entry->is_deleted = 0;
+	new_entry->min_port = min_port;
+	new_entry->max_port = max_port;
+	/***** CRITICAL SECTION START *****/
+	spin_lock(&lock);
+	for (ptr = &reservedport_list; ptr->next; ptr = ptr->next); ptr->next = new_entry;
+	spin_unlock(&lock);
+	/***** CRITICAL SECTION END *****/
+	if (min_port != max_port) printk("%sSAKURA-NOTICE: Port %u-%u reserved.\n", ccs_log_level, min_port, max_port);
+	else printk("%sSAKURA-NOTICE: Port %u reserved.\n", ccs_log_level, min_port);
 	return 0;
 }
 
@@ -66,14 +67,43 @@ int SAKURA_MayAutobind(const unsigned short int port)
 	return 0;
 }
 
-int AddReservedPortPolicy(char *data, const int is_delete)
+int AddReservedPortPolicy(char *data)
 {
 	unsigned int from, to;
+	if (!isRoot()) return -EPERM;
 	NormalizeLine(data);
 	if (sscanf(data, "%u-%u", &from, &to) == 2) {
-		if (from <= to && to < 65536) return AddReservedEntry(from, to, is_delete);
+		if (from <= to && to < 65536) return AddReservedEntry(from, to);
 	} else if (sscanf(data, "%u", &from) == 1) {
-		if (from < 65536) return AddReservedEntry(from, from, is_delete);
+		if (from < 65536) return AddReservedEntry(from, from);
+	}
+	printk("%s: ERROR: Invalid port range '%s'\n", __FUNCTION__, data);
+	return -EINVAL;
+}
+
+static int DelReservedEntry(const unsigned short int min_port, const unsigned short int max_port)
+{
+	RESERVED_ENTRY *ptr;
+	for (ptr = reservedport_list.next; ptr; ptr = ptr->next) {
+		if (ptr->min_port == min_port && max_port == ptr->max_port && !ptr->is_deleted) {
+			ptr->is_deleted = 1;
+			if (min_port != max_port) printk("SAKURA-NOTICE: Port %u-%u freed.\n", min_port, max_port);
+			else printk("SAKURA-NOTICE: Port %u freed.\n", min_port);
+			return 0;
+		}
+	}
+	return -EINVAL;
+}
+
+int DelReservedPortPolicy(char *data)
+{
+	unsigned int from, to;
+	if (!isRoot()) return -EPERM;
+	NormalizeLine(data);
+	if (sscanf(data, "%u-%u", &from, &to) == 2) {
+		if (from <= to && to < 65536) return DelReservedEntry(from, to);
+	} else if (sscanf(data, "%u", &from) == 1) {
+		if (from < 65536) return DelReservedEntry(from, from);
 	}
 	printk("%s: ERROR: Invalid port range '%s'\n", __FUNCTION__, data);
 	return -EINVAL;
@@ -87,9 +117,9 @@ int ReadReservedPortPolicy(IO_BUFFER *head)
 		head->read_var2 = (void *) ptr;
 		if (!ptr->is_deleted) {
 			if (ptr->min_port != ptr->max_port) {
-				if (io_printf(head, KEYWORD_DENY_AUTOBIND "%u-%u\n", ptr->min_port, ptr->max_port)) break;
+				if (io_printf(head, KEYWORD_DENY_AUTOBIND "%d-%d\n", ptr->min_port, ptr->max_port)) break;
 			} else {
-				if (io_printf(head, KEYWORD_DENY_AUTOBIND "%u\n", ptr->min_port)) break;
+				if (io_printf(head, KEYWORD_DENY_AUTOBIND "%d\n", ptr->min_port)) break;
 			}
 		}
 		ptr = ptr->next;
