@@ -5,7 +5,7 @@
  *
  * Copyright (C) 2005-2006  NTT DATA CORPORATION
  *
- * Version: 1.1.3   2006/07/13
+ * Version: 1.1   2006/04/01
  *
  * This file is applicable to both 2.4.30 and 2.6.11 and later.
  * See README.ccs for ChangeLog.
@@ -46,7 +46,7 @@ int DropTaskCapability(char __user * __user *args)
 	};
 	char *page;
 	if (!args) return -EINVAL;
-	page = ccs_alloc(PAGE_SIZE);
+	page = kmalloc(PAGE_SIZE, GFP_KERNEL);
 	if (!page) return -ENOMEM;
 	while (*args) {
 		int i;
@@ -69,7 +69,7 @@ int DropTaskCapability(char __user * __user *args)
 		}
 		args++;
 	}
-	ccs_free(page);
+	kfree(page);
 	return -EAGAIN;
 }
 
@@ -96,7 +96,7 @@ int CheckTaskCapability(const unsigned int operation)
 			break;
 		}
 		printk("SAKURA-ERROR: %s (pid=%d:exe=%s): Permission denied.\n", name, current->pid, exename);
-		if (exename) ccs_free(exename);
+		if (exename) kfree(exename);
 		return -EPERM;
 	}
 	return 0;
