@@ -5,7 +5,7 @@
  *
  * Copyright (C) 2005-2006  NTT DATA CORPORATION
  *
- * Version: 1.1.2   2006/06/02
+ * Version: 1.1   2006/04/01
  *
  * This file is applicable to both 2.4.30 and 2.6.11 and later.
  * See README.ccs for ChangeLog.
@@ -42,6 +42,9 @@ static int AddChrootACL(const char *entry)
 	/* I don't want to add if it was already added. */
 	for (ptr = chroot_list.next; ptr; ptr = ptr->next) if (strcmp(ptr->dir, entry) == 0) { ptr->is_deleted = 0; return 0; }
 	if ((cp = SaveName(entry)) == NULL || (new_entry = (CHROOT_ENTRY *) alloc_element(sizeof(CHROOT_ENTRY))) == NULL) goto out;
+	memset(new_entry, 0, sizeof(CHROOT_ENTRY));
+	new_entry->next = NULL;
+	new_entry->is_deleted = 0;
 	new_entry->dir = cp;
 	{
 		static spinlock_t lock = SPIN_LOCK_UNLOCKED;
@@ -76,23 +79,27 @@ int CheckChRootPermission(const char *pathname)
 		kfree(name);
 	}
 	if (error_flag) {
-		int error = -EPERM;
 		const int is_enforce = CheckCCSEnforce(CCS_SAKURA_RESTRICT_CHROOT);
 		const char *realname = realpath(pathname), *exename = GetEXE();
 		printk("SAKURA-%s: chroot %s (pid=%d:exe=%s): Permission denied.\n", GetMSG(is_enforce), realname ? realname : pathname, current->pid, exename);
-		if (is_enforce && CheckSupervisor("# %s is requesting\nchroot %s\n", exename, realname ? realname : pathname) == 0) error = 0;
 		if (exename) kfree(exename);
 		if (!is_enforce && CheckCCSAccept(CCS_SAKURA_RESTRICT_CHROOT) && realname) AddChrootACL(realname);
 		if (realname) kfree(realname);
-		if (is_enforce) return error;
+		if (is_enforce) return -EPERM;
 	}
 	return 0;
 }
 
-int AddChrootPolicy(char *data, const int is_delete)
+int AddChrootPolicy(char *data)
+{
+	if (!isRoot()) return -EPERM;
+	return AddChrootACL(data);
+}
+
+int DelChrootPolicy(const char *data)
 {
 	CHROOT_ENTRY *ptr;
-	if (!is_delete) return AddChrootACL(data);
+	if (!isRoot()) return -EPERM;
 	for (ptr = chroot_list.next; ptr; ptr = ptr->next) if (strcmp(ptr->dir, data) == 0) ptr->is_deleted = 1;
 	return 0;
 }
