@@ -5,7 +5,7 @@
  *
  * Copyright (C) 2005-2006  NTT DATA CORPORATION
  *
- * Version: 1.1.2   2006/06/02
+ * Version: 1.1   2006/04/01
  *
  * This file is applicable to both 2.4.30 and 2.6.11 and later.
  * See README.ccs for ChangeLog.
@@ -30,51 +30,45 @@ static int AuditSignalLog(const int signal, const char *dest_pattern, const int
 
 /*************************  SIGNAL ACL HANDLER  *************************/
 
-static int AddSignalEntry(const int sig, const char *dest_pattern, struct domain_info *domain, const int is_delete)
+static int AddSignalEntry(const int sig, const char *dest_pattern, struct domain_info *source)
 {
 	struct acl_info *ptr;
 	const char *saved_dest_pattern;
 	const unsigned int type_hash = MAKE_ACL_TYPE(TYPE_SIGNAL_ACL) + MAKE_ACL_HASH(sig);
 	int error = -ENOMEM;
-	if (!domain) return -EINVAL;
+	if (!source) return -EINVAL;
 	if (!dest_pattern) {
 		printk("%s: ERROR: dest_pattern == NULL\n", __FUNCTION__);
 		return -EINVAL;
 	}
 	down(&domain_acl_lock);
-	if (!is_delete) {
-		if ((ptr = domain->first_acl_ptr) == NULL) goto first_entry;
-		while (1) {
-			SIGNAL_ACL_RECORD *new_ptr;
-			if (ptr->type_hash == type_hash) {
-				if (strcmp(((SIGNAL_ACL_RECORD *) ptr)->domainname, dest_pattern) == 0) {
-					/* Found. Nothing to do. */
-					error = 0;
-					break;
-				}
+	if ((ptr = source->first_acl_ptr) == NULL) goto first_entry;
+	while (1) {
+		SIGNAL_ACL_RECORD *new_ptr;
+		if (ptr->type_hash == type_hash) {
+			const int len = strlen(((SIGNAL_ACL_RECORD *) ptr)->domainname);
+			if (strncmp(((SIGNAL_ACL_RECORD *) ptr)->domainname, dest_pattern, len) == 0 && (dest_pattern[len] == ' ' || dest_pattern[len] == '\0')) {
+				/* Found. Nothing to do. */
+				error = 0;
+				break;
 			}
-			if (ptr->next) {
-				ptr = ptr->next;
-				continue;
-			}
-		first_entry: ;
-			/* Not found. Append it to the tail. */
-			if ((saved_dest_pattern = SaveName(dest_pattern)) == NULL) break;
-			if ((new_ptr = (SIGNAL_ACL_RECORD *) alloc_element(sizeof(SIGNAL_ACL_RECORD))) == NULL) break;
-			new_ptr->type_hash = type_hash;
-			new_ptr->domainname = saved_dest_pattern;
-			error = AddDomainACL(ptr, domain, (struct acl_info *) new_ptr);
-			break;
 		}
-	} else {
-		struct acl_info *prev = NULL;
-		error = -ENOENT;
-		for (ptr = domain->first_acl_ptr; ptr; prev = ptr, ptr = ptr->next) {
-			if (type_hash != ptr->type_hash) continue;
-			if (strcmp(((SIGNAL_ACL_RECORD *) ptr)->domainname, dest_pattern)) continue;
-			error = DelDomainACL(prev, domain, ptr);
-			break;
+		if (ptr->next) {
+			ptr = ptr->next;
+			continue;
 		}
+	first_entry: ;
+		/* Not found. Append it to the tail. */
+		if ((saved_dest_pattern = SaveName(dest_pattern)) == NULL) break;
+		if ((new_ptr = (SIGNAL_ACL_RECORD *) alloc_element(sizeof(SIGNAL_ACL_RECORD))) == NULL) break;
+		new_ptr->next = NULL;
+		new_ptr->type_hash = type_hash;
+		new_ptr->domainname = saved_dest_pattern;
+		mb(); /* Instead of using spinlock. */
+		if (!ptr) source->first_acl_ptr = (struct acl_info *) new_ptr;
+		else ptr->next = (struct acl_info *) new_ptr;
+		error = 0;
+		break;
 	}
 	up(&domain_acl_lock);
 	return error;
@@ -82,7 +76,7 @@ static int AddSignalEntry(const int sig, const char *dest_pattern, struct domain
 
 int CheckSignalACL(const int sig, const int pid)
 {
-	struct domain_info *domain = current->domain_info;
+	struct domain_info *source = GetCurrentDomain();
 	struct domain_info *dest = NULL;
 	const char *dest_pattern;
 	struct acl_info *ptr;
@@ -91,7 +85,7 @@ int CheckSignalACL(const int sig, const int pid)
 	if (!CheckCCSFlags(CCS_TOMOYO_MAC_FOR_SIGNAL)) return 0;
 	if (!sig) return 0;                               /* No check for NULL signal. */
 	if (current->pid == pid) {
-		AuditSignalLog(sig, domain->domainname, 1);
+		AuditSignalLog(sig, source->domainname, 1);
 		return 0;                /* No check for self. */
 	}
 	{ /* Simplified checking. */
@@ -105,12 +99,12 @@ int CheckSignalACL(const int sig, const int pid)
 		read_unlock(&tasklist_lock);
 		if (!dest) return 0; /* I can't find destinatioin. */
 	}
-	if (domain == dest || (domain->attribute & DOMAIN_ATTRIBUTE_TRUSTED) == DOMAIN_ATTRIBUTE_TRUSTED) {
+	if (source == dest || (source->attribute & DOMAIN_ATTRIBUTE_TRUSTED) == DOMAIN_ATTRIBUTE_TRUSTED) {
 		AuditSignalLog(sig, dest->domainname, 1);
 		return 0;
 	}
 	dest_pattern = dest->domainname;
-	for (ptr = domain->first_acl_ptr; ptr; ptr = ptr->next) {
+	for (ptr = source->first_acl_ptr; ptr; ptr = ptr->next) {
 		if (ptr->type_hash == type_hash) {
 			const int len = strlen(((SIGNAL_ACL_RECORD *) ptr)->domainname);
 			if (strncmp(((SIGNAL_ACL_RECORD *) ptr)->domainname, dest_pattern, len) == 0 && (dest_pattern[len] == ' ' || dest_pattern[len] == '\0')) break;
@@ -121,21 +115,22 @@ int CheckSignalACL(const int sig, const int pid)
 		return 0;
 	}
 	if (TomoyoVerboseMode()) {
-		printk("TOMOYO-%s: Signal %d to %s denied for %s\n", GetMSG(is_enforce), sig, GetLastName(dest), GetLastName(domain));
+		printk("TOMOYO-%s: Signal %d to %s denied for %s\n", GetMSG(is_enforce), sig, GetLastName(dest), GetLastName(source));
 	}
 	AuditSignalLog(sig, dest_pattern, 0);
-	if (is_enforce) return CheckSupervisor("%s\n" KEYWORD_ALLOW_SIGNAL "%d %s\n", domain->domainname, sig, dest_pattern);
-	if (CheckCCSAccept(CCS_TOMOYO_MAC_FOR_SIGNAL)) AddSignalEntry(sig, dest_pattern, domain, 0);
+	if (is_enforce) return -EPERM;
+	if (CheckCCSAccept(CCS_TOMOYO_MAC_FOR_SIGNAL)) AddSignalEntry(sig, dest_pattern, source);
 	return 0;
 }
 
 
-int AddSignalPolicy(char *data, struct domain_info *domain, const int is_delete)
+int AddSignalPolicy(char *data, void **domain)
 {
 	int sig;
-	char *domainname = strchr(data, ' ');
-	if (sscanf(data, "%d", &sig) == 1 && domainname && IsDomainDef(domainname + 1)) {
-		return AddSignalEntry(sig, domainname + 1, domain, is_delete);
+	char *cp;
+	if (!isRoot()) return -EPERM;
+	if (sscanf(data, "%d", &sig) == 1 && (cp = strchr(data, ' ')) != NULL) {
+		if (IsDomainDef(cp + 1)) return AddSignalEntry(sig, cp + 1, (struct domain_info *) *domain);
 	}
 	return -EINVAL;
 }
@@ -154,7 +149,8 @@ static int AuditSignalLog(const int signal, const char *dest_pattern, const int
 	if (CanSaveAuditLog(is_granted) < 0) return -ENOMEM;
 	do_gettimeofday(&tv);
 	len = strlen(domainname) + strlen(dest_pattern) + 256;
-	if ((buf = kzalloc(len, GFP_KERNEL)) == NULL) return -ENOMEM;
+	if ((buf = kmalloc(len, GFP_KERNEL)) == NULL) return -ENOMEM;
+	memset(buf, 0, len);
 	snprintf(buf, len - 1, "#timestamp=%lu pid=%d uid=%d gid=%d euid=%d egid=%d suid=%d sgid=%d fsuid=%d fsgid=%d\n%s\n" KEYWORD_ALLOW_SIGNAL "%d %s\n", tv.tv_sec, task->pid, task->uid, task->gid, task->euid, task->egid, task->suid, task->sgid, task->fsuid, task->fsgid, domainname, signal, dest_pattern);
 	return WriteAuditLog(buf, is_granted);
 }
